#!/bin/bash
#
# Script for running SCAFES OR SCADOPT applications.
#
# Some useful commands for LSF systems.
# All pending jobs will be killed:
#     bkill ` bjobs -u <user> |grep PEND |cut -f1 -d" "`
# Reserverations will be shown:
#     brsrv
#
# Create on own result directory for run with and without taskset.
#
# Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
# For details, see the files COPYING and LICENSE in the base directory
# of the package.

set -u

#------------------------------------------------------------------------------#
declare    local nameProblem=${SCADOPT_NAME_EXECUTABLE}
declare -i local spaceDim=$SCADOPT_SPACE_DIM;
declare -i local nTestsMpi=$SCADOPT_N_TESTS_MPI;
declare -i local nTestsOpenMp=$SCADOPT_N_TESTS_OPENMP;
declare    local testForWeakScaling=$SCADOPT_TYPE_SCALINGTEST;
declare -i local nProcessesMpiStart=$SCADOPT_N_PROCESSES_MPI_START;
declare -i local nThreadsOpenMpStart=$SCADOPT_N_THREADS_OPENMP_START;
declare -i local nTimesteps=$SCADOPT_N_TIMESTEPS;
declare -i local nSnapshots=$SCADOPT_N_SNAPSHOTS;
declare -i local starttime=$SCADOPT_START_TIME;
declare -i local endtime=$SCADOPT_END_TIME; #TODO: Must be set by hand.
declare    local nameConfigfile=$SCADOPT_NAME_CONFIGFILE;
declare    local nameKindfile=$SCADOPT_NAME_KINDFILE;
declare    local writeKindfile=$SCADOPT_WRITE_KINDFILE;
declare    local enableAdolc=$SCADOPT_ENABLE_ADOLC;
declare    local asynchronMode=$SCADOPT_ASYNCHRON_MODE;
declare    local computeGradients=$SCADOPT_COMP_GRADIENTS;
declare    local isTfTimeDependent=$SCADOPT_TF_TIME_DEPENDENT;
declare    local runMode=$SCADOPT_RUN_MODE;
declare    local outputlevel=$SCADOPT_OUTPUT_LEVEL;
declare -i local oneAsInteger=1;
declare -i local nNodesMachineMAX=$SCADOPT_MACHINE_N_NODES;
declare -i local nCoresPerNodeMAX=$SCADOPT_MACHINE_N_CORES_PER_NODE;
declare    local nameMachine=${SCADOPT_MACHINE_NAME};
declare    local softwareVersion="ScaFES_2_0_0";
declare    local nLayersAtBorder="${SCADOPT_NLAYERS_AT_BORDER}";
declare    local resDir=${SCADOPT_RESULTS_DIR};
declare    local checkResults=${SCADOPT_CHECK_RESULTS};

#------------------------------------------------------------------------------#
# Independent internal variables.
declare local nameReservation="hpc-flis";
declare local nameProject="hpc-flis";
declare local memoryPerCore="300";

#------------------------------------------------------------------------------#
# Compute value of dependent variable.
declare -i local nCoresTotalMAX=${nNodesMachineMAX};
nCoresTotalMAX=$((nCoresTotalMAX*nCoresPerNodeMAX));

#------------------------------------------------------------------------------#
if [ "$runMode" == "DEBUG" ] ; then
    nTestsMpi=1;
    nProcessesMpiStart=1;
    #nThreadsOpenMpStart=1;
fi
#------------------------------------------------------------------------------#
if [ "$nameKindfile" == "" ] ; then
    optionKindfile="";
else
    optionKindfile="--kindfile=$nameKindfile";
fi
#------------------------------------------------------------------------------#
if [ "$nameConfigfile" == "" ] ; then
    optionConfigfile="";
else
    optionConfigfile="--configfile=$nameConfigfile";
fi
#------------------------------------------------------------------------------#
boolEnabledAdolc=0;
if [ "$enableAdolc" == "YES" ] ; then
    boolEnabledAdolc=1;
fi
boolAsynchronMode=0;
if [ "$asynchronMode" == "YES" ] ; then
    boolAsynchronMode=1;
fi
boolCompGradients=0;
if [ "$computeGradients" == "YES" ] ; then
    boolCompGradients=1;
fi
boolIsTfTimeDependent=0;
if [ "$isTfTimeDependent" == "YES" ] ; then
    boolIsTfTimeDependent=1;
fi

boolWriteKindfile=0;
if [ "$writeKindfile" == "YES" ] ; then
    boolWriteKindfile=1;
fi


declare -i local endTestMpi;
declare -i local endTestOpenMp;
declare    local currNameProblem;
declare -i local currIdxProblem;
declare -i local currTest;
declare -i local currIdTest;
declare -i local currIdTestMpi;
declare -i local currIdTestOpenMp;
declare -i local currNprocessesMpi;
declare -i local currNthreadsOpenMp;
declare -i local currEndTest;
declare    local currNameExecutable; # Relative path to executable
declare -i tmpNodesDir;
declare    local errCode=0;
endTestMpi=${nTestsMpi}-${oneAsInteger};
endTestOpenMp=${nTestsOpenMp}-${oneAsInteger};
currEndTest=${nTestsMpi}*${nTestsOpenMp};

timestamp=$(date +%Y%m%d%H%M);
declare local nNodesMachine=1;

currNameProblem=${nameProblem};
currNameExecutable="./${currNameProblem}";
currIdTest=$oneAsInteger;
(
#------------------------------------------------------------------------------#
# Perform a loop over all MPI tests.
currNprocessesMpi=${nProcessesMpiStart};
for idxTestMpi in `seq 0 $endTestMpi`; do
        currIdTestMpi=${idxTestMpi}+${oneAsInteger};

        #----------------------------------------------------------------------#
        # Perform a loop over all OpenMP tests.
        currNthreadsOpenMp=${nThreadsOpenMpStart};
        for idxTestOpenMp in `seq 0 $endTestOpenMp`; do
            currIdTestOpenMp=${idxTestOpenMp}+${oneAsInteger};
            #currNnodes=(${SCADOPT_N_NODES[*]})
            currStarttime=${starttime};
            currEndtime=${endtime};
            currNtimesteps=${nTimesteps};
            currNsnapshots=${nSnapshots};

            #------------------------------------------------------------------#
            divideGrid=${SCADOPT_PARTITION_GRID};
            currNnodes="${SCADOPT_N_NODES}";
            currCoordNodeFirst="${SCADOPT_COORD_NODE_FIRST}";
            currCoordNodeLast="${SCADOPT_COORD_NODE_LAST}";
            namePartfile="${currNameProblem}_${currNprocessesMpi}";
            namePartfile="${namePartfile}_${currNthreadsOpenMp}_${currNnodes}";
            nameDatafile=${namePartfile};

            #------------------------------------------------------------------#
            str_currNprocessesMpi=$(printf "%.5d" $currNprocessesMpi);
            str_currNthreadsOpenMp=$(printf "%.5d" $currNthreadsOpenMp);
            str_nTimesteps=$(printf "%.6d" $currNtimesteps);
            basefile="${resDir}/${currNameProblem}"
            basefile="$basefile-${nameMachine}"
            basefile="$basefile-${softwareVersion}"
            basefile="$basefile-mpi${str_currNprocessesMpi}"
            basefile="$basefile-omp${str_currNthreadsOpenMp}"
            basefile="$basefile-nNodes${currNnodes}"
            basefile="$basefile-divide${divideGrid}"
            basefile="$basefile-adolc${enableAdolc}"
            basefile="$basefile-async${asynchronMode}"
            basefile="$basefile-grad${computeGradients}"
            basefile="$basefile-tfTdep${isTfTimeDependent}"
            basefile="$basefile-nTs${str_nTimesteps}"
            basefile="$basefile-tEnd${endtime}"
            basefile="$basefile-nLayers${nLayersAtBorder}"
            basefile="$basefile-${timestamp}"
            outputfile="$basefile.out"
            errorfile="$basefile.err"

            #------------------------------------------------------------------#
            if ( expr ${currNprocessesMpi} \< ${nNodesMachineMAX} > /dev/null ) ; then
                nNodesMachine=1;
            else
                nNodesMachine=$((currNprocessesMpi/nNodesMachineMAX));
            fi

            declare -i local nCoresTotal=$((currNprocessesMpi*currNthreadsOpenMp));
            if [ ${nCoresTotal} -gt ${nCoresTotalMAX} ]; then
               continue;
            fi

            echo -e "\n\n  +-------------------------------+";
            echo -e "  | TEST: ${currIdTest} / ${currEndTest}";
            printf "  |     * Name of Problem     = %s\n" ${currNameProblem};
            printf "  |     * #{MPI processes}    = %s\n" ${currNprocessesMpi};
            printf "  |     * #{OpenMP threads}   = %s\n" ${currNthreadsOpenMp};
            printf "  |     * #{Nodes}            = %s\n" ${currNnodes};
            printf "  |     * divideGrid          = %s\n" ${divideGrid};
            printf "  |     * enableAdolc         = %s\n" ${enableAdolc};
            printf "  |     * asynchronMode       = %s\n" ${asynchronMode};
            printf "  |     * computeGradients    = %s\n" ${computeGradients};
            printf "  |     * isTfTimeDependent   = %s\n" ${isTfTimeDependent};
            printf "  |     * #{Start time}       = %s\n" ${currStarttime};
            printf "  |     * #{End time}         = %s\n" ${currEndtime};
            printf "  |     * #{Time steps}       = %s\n" ${currNtimesteps};
            printf "  |     * #{Layers At Border} = %s\n" ${nLayersAtBorder};
            printf "  |     * #{Snap shots}       = %s\n  |\n" ${currNsnapshots};
            printf "  |     * #{NodesMachineUsed} = %s\n" ${nNodesMachine};
            printf "  |     * #{CoresTotalUsed}   = %s\n  |\n" \
                        $((currNprocessesMpi*currNthreadsOpenMp));

            printf "  |     * Name of Machine     = %s\n" ${nameMachine};
            printf "  |     * #{NodesMachineMAX}  = %s\n" ${nNodesMachineMAX};
            printf "  |     * #{CoresPerNodeMAX}  = %s\n" ${nCoresPerNodeMAX};
            printf "  |     * #{CoresTotalMAX}    = %s\n" ${nCoresTotalMAX};
            echo -e "  +-------------------------------+\n\n";

            #------------------------------------------------------------------#
            echo "--- Run application. ---"
            exeString="$currNameExecutable \
                    --dim=$spaceDim \
                    --divideGrid=$divideGrid \
                    --nNodes=$currNnodes \
                    --coordNodeFirst=$currCoordNodeFirst \
                    --coordNodeLast=$currCoordNodeLast \
                    --starttime=$currStarttime \
                    --endtime=$currEndtime \
                    --nTimesteps=$currNtimesteps \
                    --nSnapshots=$currNsnapshots \
                    ${optionKindfile} \
		    --writeKindFile=$boolWriteKindfile \
                    ${optionConfigfile} \
                    --enabledAdolc=$boolEnabledAdolc \
                    --asynchronMode=$boolAsynchronMode \
                    --computeGradients=$boolCompGradients \
                    --isTfTimeDependent=$boolIsTfTimeDependent \
                    --writePartitionfile=$namePartfile \
                    --outputfile=$nameDatafile \
                    --outputlevel=$outputlevel \
		    --nLayersAtBorder=$nLayersAtBorder"

            case $runMode in
            #------------------------------------------------------------------#
            DEBUG )
                (
                set -x
                export OMP_NUM_THREADS=$currNthreadsOpenMp;
                libtool --mode=execute gdb --args $exeString
                )
                ;;
            #------------------------------------------------------------------#
            NORMAL )
                (
                set -x
                export OMP_NUM_THREADS=$currNthreadsOpenMp;
                /usr/bin/time mpirun -np $currNprocessesMpi \
                    $exeString > ${outputfile} 2> ${errorfile}
                )
                ;;
            #------------------------------------------------------------------#
            LSF )
                (
                # -npernode 64
                set -x
                export OMP_NUM_THREADS=$currNthreadsOpenMp;
                #  -U ${nameReservation}
                bsub \
                  -P ${nameProject} \
                  -n ${nCoresTotal} \
                  -o ${outputfile} \
                  -e ${errorfile} \
                  -J ${currNameExecutable} \
                  -M ${memoryPerCore} \
                  mpirun -np $currNprocessesMpi \
                    $exeString > ${outputfile} 2> ${errorfile}
                )
                ;;
            #------------------------------------------------------------------#
            SLURM )
                (
                set -x
                srun -v -v \
                    --nodes=${nNodesMachine} \
                    --ntasks=${currNprocessesMpi} \
                    --reservation=${nameReservation} \
                    -o ${outputfile} \
                    -e ${errorfile} \
                    -J ${currNameExecutable} \
                    -A ${nameProject} \
                    --time=01:00:00 \
                    --mem-per-cpu=${memoryPerCore} \
                    --partition=sandy \
                    $exeString > ${outputfile} 2> ${errorfile}
                )
                ;;
            #------------------------------------------------------------------#
            * )
                echo "ERROR: Mode not known." ;;
            esac
            errCode=$?
            echo -e "";
            if [ "${errCode}" != "0" ]; then
                echo -e "RUN TIME ERROR CODE = $errCode. Exit!";
                echo -e "RUN TIME ERROR = YES";
                break 2;
            else
                echo -e "* RUN SUCCESSFULLY?    YES";
            fi
            #------------------------------------------------------------------#
            if [ "$checkResults" == "YES" ]; then
                grep -h "Local Linf" ${outputfile} > tmp.dat
                correctErrCode=$?
                if [ "$correctErrCode" != "0" ]; then
                    echo "CORRECTNESS CHECK: FILE NOT FOUND. Exit!";
                    errCode=1;
                    break 2;
                fi

                grep -h -v "0.0000" tmp.dat
                correctErrCode=$?
                # If lines with error!=0 were found.
                if [ "$correctErrCode" == "0" ]; then
                    echo "* ARE RESULTS CORRECT? NO";
                    errCode=1;
                    break 2;
                else
                    echo "* ARE RESULTS CORRECT? YES";
                    errCode=0;
                fi
            else
                echo "* ARE RESULTS CORRECT? UNKNOWN (DID NOT CHECK)";
            fi

            currNthreadsOpenMp=$currNthreadsOpenMp*2;
            currIdTest=$currIdTest+$oneAsInteger;
        done

        #------------------------------------------------------------------#
        # Test for weak scalability:
        # Problem size increases linearly with number of processes.
        #------------------------------------------------------------------#
        # Test for strong scalability:
        # Problem size remains constant.
        # The number of processes increases linearly.
        if [ "$testForWeakScaling" == "WEAK" ]; then
            echo "TODO: nNodesTotal has to be doubled!";
        else
            currNprocessesMpi=$currNprocessesMpi*2;
        fi
done
exit $errCode;
)
