/* ScaFES
 * Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
 * For details, see the files COPYING and LICENSE in the base directory
 * of the package.
 */

/**
 *  @file ScaFES.hpp
 *
 *  @brief Main include file for applications using ScaFES.
 *
 * This file has to be included if a new initial boundary value problem
 * should be created.
 */

/**
 *  \mainpage ScaFES Documentation
 *
 * "We present ScaFES, an open-source HPC framework written in C++11 for
 * solving initial boundary value problems using explicit numerical methods
 * in time on structured grids. It is designed to be highly-scalable and
 * very user-friendly, i.e. to exploit all levels of parallelism and provide
 * easy-to-use interfaces. Besides, the numerical nomenclature is reflected
 * in a nearly one-to-one mapping."
 *
 * [Source: Martin Flehmig, Kim Feldhoff, and Ulf Markwardt:
 * "ScaFES: An Open-Source Framework for Explcit Solvers Combing
 * High-Scalability with User-Friendliness",
 * ARCS 2014 - 27th International Conference of Computing Systems,
 * Workshop-Proceedings, VDE-Verlag, 2014]
 *
 * The framework is freely available under the new BSD license. For details,
 * please refer to the files COPYING and LICENSE in the base directory
 * of the package.
 *
 * \section Implementation How to Implement a New Problem?
 * The usual way to implement a new discretised problem is as follows:
 * - Create the skeleton of the discretized problem.
 * The skeleton of the discretised problem can be setup by creating a new
 * class which is derived from the class \c ScaFES::Problem.
 * - Add all fields to the problem.
 * All fields can be added to the discrete problem via the constructor
 * of the base class \c ScaFES::Problem.
 * - Evaluate all fields at all grid nodes.
 * All known fields can be evaluated at all global inner resp. border grid nodes
 * at all time steps by implementing the methods \c evalInner resp.
 * \c evalBorder.
 * - Initialise all unknown fields at all grid nodes.
 * All unknown fields can be initialised at all global inner resp. border
 * grid nodes at the starting time by implementing the methods \c initInner
 * resp. \c initBorder.
 * -  Update all unknown fields at all grid nodes.
 * All unknown fields can be updated at all global inner resp. border
 * grid nodes at the starting time by implementing the methods \c updateInner
 * resp. \c updateBorder.
 * - Write the main program.
 *
 * \section example A First Example
 * In the following, we show how a new time-dependent PDE problem can be
 * implemented in ScaFES. Fur this purpose, we have chosen the 3-D heat equation
 * on the unit cube as this example serves as a prototype for the considered
 * problem class.
 *
 * \ref  heatEqnFDM "Discretized heat equation problem on unit cube"
 *
 * \section contact Contact
 * Bugs can be reported to: kim.feldhoff@tu-dresden.de
 *
 * More information about ScaFES like new releases can be found at:
 * <a href="http://www.tu-dresden.de/zih/scafes">ScaFES web page</a>
 *
 * \section funding Funding
 * ScaFES was funded by the Federal Ministry of Education and Research (BMBF)
 * within the project HPC-FLiS under the support code 01 IH 11 009.
 *
 ******************************************************************************/

#ifndef SCAFES_HPP_
#define SCAFES_HPP_

#include "ScaFES_Config.hpp"

#include "ScaFES_Problem.hpp"

#endif
